''**************************
''* Using assembly code    *
''* as threads in a Cog    *
''* (C) 2007 Thomas Kliche *
''**************************
''
''This is a multithread kernel which allows you to use different assembly routines running
''simultaneously in the same Cog. This can be used to have some low speed drivers running in one
''Cog. 
'' 
''This code is based on an idea of Bill Henning which has built a virtual processor by a few
''assembly instructions to use the main memory for assembly code. This is called large
''memory model. For more details please refer
''http://forums.parallax.com/forums/default.aspx?f=25&p=1&m=154421
''I have modified the code of the virtual processor so that it can only execute assembly code
''which is located in Cog memory.
''
''If you have some code or if you want to develop new code which you want to use as a thread
''running, you have to take care on the following limitations.
''
''1. djnz, jmp, jmpret and call instructions
''
''   You cannot use djnz, jmp/jmpret and call instructions in the ordinary manner because the code
''   is executed in a virtual processor and you would leave these with such an instruction. You
''   can use this instructions only to call routines in the kernel. Otherwise use the following
''   code as replacement.
''
''   a) djnz replacement
''             sub    counter,#1 wz     'equivalent to djnz counter,#loop
''      if_nz  mov    PC,#loop
''
''   b) jmp replacement
''             mov    PC,#label         'equivalent to jmp #label
''
''   c) call replacement
''             mov    CallAddr,#label
''             jmpret label_ret,#vCall  'equivalent to call #label 
''                                      'label on ret must be specify manually
''
''   d) jmpret replacement
''             mov    CallAddr,#label
''             jmpret retVar,#vJmpRet   'equivalent to jmpret retVar,#label
''
''2. waitcnt, waitpeq, waitpne and waitvid instructions
''
''   You should not use this instructions because the Cog will pause and therefore all threads
''   would be pause. You can use the following code as replacement.
''
''   a) waitcnt replacement
''      mov    Wait,#count
''      jmp    #vWaitCnt         'equivalent to waitcnt _SysCnt,#count
''
''      Please take care on the following limitation. Depending on the count of used threads the
''      minimum count for wait can be a multiple of _CLOCKS_PER_TIMESLICE. This will be calculated
''      when you call vWaitCnt. If your value for wait is lower than the calculated value then
''      the calculated value will be used instead. E.g. if you have 4 threads running at the
''      default timeslice at 10 ms then the lowest wait count is 30 ms.
''
''   b) waitpeq and waitpne instructions
''      Currently I'm working on a way to have a similar function in kernel. Until this is done
''      you will not be able to use this instructions without the limitation I have mentioned
''      above. 
''
''   c) waitvid instruction
''      There is no workaround possible so I prefer you do not use this instruction in a
''      threaded Cog. 
''
''
''DO NOT USE LABELS OF KERNEL BEGINNING WITH UNDERSCORE!
''THE ADDRESS OF THESE CAN BE CHANGED OR THE LABEL CAN BE REMOVED IN A FURTHER VERSION.
''
CON

  _clkmode = xtal1 + pll8x                              'demo runs at 40 MHz to show that all threads are running
  _xinfreq = 5_000_000                                  'but you can use this kernel at your preferred clock rate

  FLAG_C              = %01                             'bit position for C flag in variable Flags
  FLAG_NZ             = %10                             'bit position for inverted Z flag in variable Flags
  REGISTER_SIZE       = 0                               'size of registers saved with context
  VARIABLE_SIZE       = 5                               'size of variables saved with context
  CONTEXT_ENTRY_SIZE  = REGISTER_SIZE + VARIABLE_SIZE   'size of context entry
  CLOCKS_OF_YIELD     = 200                             'clocks needed by yield routine

  
PUB start

  cognew(@entry, 0)


DAT

'requires 91 longs + space for thread context table with a size of minimum 5 longs per entry
entry                   org     0

vNext                   jmp     #_v_next                'process next instruction
vCall                   jmp     #_v_call                'call a subroutine
vJmpRet                 jmp     #_v_jmpret              'call a dynamic subroutine
vYield                  jmp     #_v_yield               'switch to the next thread
vWaitCnt                jmp     #_v_wait                'pause the current thread for a specified time
                        jmp     #_v_next                'reserved for vWaitPEQ
                        jmp     #_v_next                'reserved for vWaitPNE
                        jmp     #_v_next

'following variables are saved in thread context
_SysCnt                 long    0                       'system counter used for wait
PC                      long    CallAddr                'program counter of virtual processor
Flags                   long    0                       'storage to save flags, currently only bit 1 and 0 used
CallAddr                jmp     #_start                 'address of subroutine to call
Wait                    long    0                       'pause thread until clock count
'following registers are saved in thread context
'this registers can be used in a thread if the same code is used in more then one thread at a time  
{
R0                      long    0
R1                      long    0
R2                      long    0
R3                      long    0
}

'-----------------------
'virtual processor                        
_newThread              mov     _InstrCounter,_INSTR_PER_TIMESLICE                                
_nextInstr              movs    _getInstr,PC
                        add     PC,#1
_getInstr               mov     _virtInstr,0-0          'get instruction addressed by PC
                        nop
_virtInstr              nop
_v_next                 djnz    _InstrCounter,#_nextInstr
_v_yield                '196 clocks
                        muxnz   Flags,#FLAG_NZ                                                                  
                        muxc    Flags,#FLAG_C                                                                  
_yield                  movs    :ctxsave,#_SysCnt
                        movd    :ctxsave,_CurrentTask
                        mov     _InstrCounter,#CONTEXT_ENTRY_SIZE
:ctxsave                mov     0-0,0-0
                        add     :ctxsave,_INC_DEST_SRC
                        djnz    _InstrCounter,#:ctxsave

_checkNextCtx           add     _CurrentTask,#CONTEXT_ENTRY_SIZE
                        cmp     _CurrentTask,#_ThreadContext_End wz
        if_e            mov     _CurrentTask,#_ThreadContext

_start                  movs    :ctxSysCnt,_CurrentTask
                        movd    :ctxSysCnt,#_SysCnt
                        mov     _InstrCounter,#CONTEXT_ENTRY_SIZE-1 wc 'clear C flag
:ctxSysCnt              mov     0-0,0-0         wz      'exist system count to start this thread?
                        add     :ctxSysCnt,_INC_DEST_SRC
                        mov     :ctxload,:ctxSysCnt
        if_nz           cmps    _SysCnt,cnt        wc   'if syscnt != 0, is time to wait left?
        if_nc_and_nz    jmp     #_checkNextCtx          'proceed next thread
        if_nz           mov     _SysCnt,#0              'reset counter
:ctxload                mov     0-0,0-0
                        add     :ctxload,_INC_DEST_SRC
                        djnz    _InstrCounter,#:ctxload
                        cmp     PC,#0           wz      'is this thread ready?
        if_e            jmp     #_checkNextCtx          'no, then check next context
                        shr     Flags,#1        nr,wc,wz'recover flags            
                        jmp     #_newThread                        

'-----------------------
                        '60 clocks + 16 per thread
_v_wait                 muxnz   Flags,#FLAG_NZ                                                                  
                        muxc    Flags,#FLAG_C
                        cmp     Wait,#0         wz
        if_e            jmp     #_yield                 'no wait is given, then it's like yield        
                        mov     _Temp,_CLOCKS_PER_TIMESLICE
                        mov     _SysCnt,#0                                                                  
                        mov     _InstrCounter,#_ThreadContext
                        jmp     #:calcThreads
                        
:minWait                add     _SysCnt,_Temp
:calcThreads            add     _InstrCounter,#CONTEXT_ENTRY_SIZE
                        cmp     _InstrCounter,#_ThreadContext_End wc
        if_b            jmp     #:minWait               '_Temp = (max. threads - 1) * clocks of timeslice 
                        'sub     Wait,#76
                        min     Wait,_SysCnt            'determine that wait is not lower than minimal value
                        mov     _SysCnt,Wait
                        add     _SysCnt,cnt             'calculate system counter
                        jmp     #_yield

'-----------------------                        
_v_call                 call    #_jmpret_call
                        movd    :retInstr,_Temp
                        movs    _RETURN,PC              'use correct return address
                        mov     PC,CallAddr             'proceed on subroutine
:retInstr               mov     0-0,_RETURN             'modify ret instruction
                        jmp     #_v_next                                              

'-----------------------                        
_v_jmpret               call    #_jmpret_call
                        movd    :jmpretVar,_Temp
                        mov     PC,CallAddr             'proceed on subroutine
:jmpretVar              mov     0-0,PC                  'store PC to user defined variable
                        jmp     #_v_next

'-----------------------                        
_jmpret_call            sub     PC,#1                   'point to jmpret
                        movs    :jmpretInstr,PC
                        add     PC,#1                        
:jmpretInstr            mov     _Temp,0-0               'get jmpret instruction
                        shr     _Temp,#9                'extract pointer to ret or variable
_jmpret_call_ret        ret
                        
'#######################################################                        
'constants used by virtual processor                        
                        'to calculate other clock count for timeslice use this formula
                        'x = (ts - 0,0025) / 0,0003
                        'where ts is defered timeslice in milliseconds  
_INSTR_PER_TIMESLICE    long    33_325                  '10 ms per thread using up to 4 threads
                                '16_658                 'or 5 ms per thread using 5 up to 8 threads
_CLOCKS_PER_TIMESLICE   long    800_000                 '399_992 for 5 ms                                
_INC_DEST_SRC           long    1 << 9 | 1
_RETURN                 mov     PC,#0                   'used to replace a ret instruction

'variables used by virtual processor
_Temp                   long    0
_InstrCounter           long    2
_CurrentTask            long    _ThreadContext
                        'system count, program counter, flags, parameter variables    
_ThreadContext          long    0,thread1,0[3]          'thread 1
                        long    0,thread2,0[3]          'thread 2
                        long    0,thread3,0[3]          'thread 3
                        long    0,thread4,0[3]          'thread 4                        
                        long    0,thread5,0[3]          'thread 5
                        long    0,thread6,0[3]          'thread 6
                        long    0,thread7,0[3]          'thread 7
                        long    0,thread8,0[3]          'thread 8                        
_ThreadContext_End
'==================================================================================================

''Sample code that demonstrate how the functions of kernel can be used. 

thread1                 mov     WaitCounter1,cnt
                        add     WaitCounter1,BlinkFreq
                        mov     dira,LEDS
:toggleLED              xor     outa,LED20
                        mov     CallAddr,#wait1         'specify call address 
                        jmpret  wait1_ret,#vCall        'is equivalent to call #wait1
                                                        'the return label must be specified manually                        
                        mov     PC,#:toggleLED          'equivalent to jmp #:toggleLED

wait1                   cmps    WaitCounter1,cnt wc
        if_nc           mov     PC,#wait1               'jmp #wait1
                        add     WaitCounter1,BlinkFreq
wait1_ret               ret


'-----------------------                        
thread2                                          
:toggleLED              xor     outa,LED21
                        jmp     #vYield                 'give next thread a chance to run
                        mov     PC,#:toggleLED          'jmp #:toggleLED


'-----------------------                        
thread3
:toggleLED              xor     outa,LED22
                        mov     Wait,#0                 'if you specify not to wait
                        jmp     #vWaitCnt               'vWaitCnt is treated like vYield
                        mov     PC,#:toggleLED          'jmp #:toggleLED


'-----------------------                        
thread4
:toggleLED              xor     outa,LED23
                        mov     Wait,BlinkFreq          'specify delta like on assembly instruction
                        jmp     #vWaitCnt               'equivalent to waitcnt tmp,BlinkFreq 
                        mov     PC,#:toggleLED          'jmp #:toggleLED


'-----------------------                        
thread5                                          
:toggleLED              xor     outa,LED19
                        jmp     #vYield                 'give next thread a chance to run
                        mov     PC,#:toggleLED          'jmp #:toggleLED


'-----------------------                        
thread6
:toggleLED              xor     outa,LED18
                        mov     Wait,BlinkFreq4
                        jmp     #vWaitCnt               'waitcnt tmp,BlinkFreq4
                        mov     PC,#:toggleLED          'jmp #:toggleLED


'-----------------------                        
thread7
:toggleLED              xor     outa,LED17
                        mov     Wait,BlinkFreq2
                        jmp     #vWaitcnt               'waitcnt tmp,BlinkFreq2 
                        mov     PC,#:toggleLED          'jmp #:toggleLED


'-----------------------                        
thread8
:toggleLED              xor     outa,LED16
                        mov     Wait,BlinkFreq          
                        jmp     #vWaitcnt               'waitcnt tmp,BlinkFreq 
                        mov     PC,#:toggleLED          'jmp #:toggleLED


LEDS                    long    $00FF_0000
LED16                   long    $0001_0000
LED17                   long    $0002_0000
LED18                   long    $0004_0000
LED19                   long    $0008_0000
LED20                   long    $0010_0000
LED21                   long    $0020_0000
LED22                   long    $0040_0000
LED23                   long    $0080_0000
BlinkFreq               long    40_000_000
BlinkFreq2              long    20_000_000
BlinkFreq4              long    10_000_000
BlinkFreq8              long    8_000_000
BlinkFreq16             long    4_000_000
WaitCounter1            res